/* -------------------------------------------- */
import { Imperium5Combat } from "./imperium5-combat.js";
import { Imperium5Commands } from "./imperium5-commands.js";

/* -------------------------------------------- */
const __level2Dice = ["d0", "d4", "d6", "d8", "d10", "d12"];
const __name2DiceValue = { "0": 0, "d0": 0, "d4": 4, "d6": 6, "d8": 8, "d10": 10, "d12": 12 }

/* -------------------------------------------- */
export class Imperium5Utility {


  /* -------------------------------------------- */
  static async init() {
    Hooks.on('renderChatLog', (log, html, data) => Imperium5Utility.chatListeners(html));
    
    Hooks.on("getCombatTrackerEntryContext", (html, options) => {
      Imperium5Utility.pushInitiativeOptions(html, options);
    })

    this.rollDataStore = {}
    this.defenderStore = {}
    this.diceList = []
    Imperium5Commands.init();

    Handlebars.registerHelper('count', function (list) {
      return list.length;
    })
    Handlebars.registerHelper('includes', function (array, val) {
      return array.includes(val);
    })
    Handlebars.registerHelper('upper', function (text) {
      return text.toUpperCase();
    })
    Handlebars.registerHelper('lower', function (text) {
      return text.toLowerCase()
    })
    Handlebars.registerHelper('upperFirst', function (text) {
      if (typeof text !== 'string') return text
      return text.charAt(0).toUpperCase() + text.slice(1)
    })
    Handlebars.registerHelper('notEmpty', function (list) {
      return list.length > 0;
    })
    Handlebars.registerHelper('mul', function (a, b) {
      return parseInt(a) * parseInt(b);
    })
    Handlebars.registerHelper('exists', function (val) {
      return val != null && val != undefined;
    })
  }

  /* -------------------------------------------- */
  static pushInitiativeOptions(html, options) {
  }

  /* -------------------------------------------- */
  static async ready() {

  }

  /* -------------------------------------------- */
  static async addItemDropToActor(actor, item) {
    actor.preprocessItem("none", item, false)
    let chatData = {
      user: game.user.id,
      rollMode: game.settings.get("core", "rollMode"),
      whisper: [game.user.id].concat(ChatMessage.getWhisperRecipients('GM')),
      content: `<div>The item ${item.name} has been dropped on the actor ${actor.name}</div`
    }
    ChatMessage.create(chatData);
  }

  /* -------------------------------------------- */
  static async loadCompendiumData(compendium) {
    const pack = game.packs.get(compendium);
    return await pack?.getDocuments() ?? [];
  }

  /* -------------------------------------------- */
  static async loadCompendium(compendium, filter = item => true) {
    let compendiumData = await Imperium5Utility.loadCompendiumData(compendium);
    return compendiumData.filter(filter);
  }

  /* -------------------------------------------- */
  static getOptionsStatusList() {
    return this.optionsStatusList;
  }

  /* -------------------------------------------- */
  static async chatListeners(html) {

    html.on("click", '.view-item-from-chat', event => {
      game.system.pegasus.creator.openItemView(event)
    })
  }

  /* -------------------------------------------- */
  static async preloadHandlebarsTemplates() {

    const templatePaths = [
      'systems/fvtt-imperium5/templates/editor-notes-gm.html',
      'systems/fvtt-imperium5/templates/actor-partial-ames.html',
      'systems/fvtt-imperium5/templates/actor-partial-paradigmes.html',
      'systems/fvtt-imperium5/templates/partial-item-description.html',
    ]
    return loadTemplates(templatePaths);
  }

  /* -------------------------------------------- */
  static async getEffectFromCompendium(effectName) {
    effectName = effectName.toLowerCase()
    let effect = game.items.contents.find(item => item.type == 'effect' && item.name.toLowerCase() == effectName)
    if (!effect) {
      let effects = await this.loadCompendium('fvtt-pegasus.effect', item => item.name.toLowerCase() == effectName)
      let objs = effects.map(i => i.toObject())
      effect = objs[0]
    } else {
      effect = duplicate(effect);
    }

    console.log("Effect", effect)
    return effect
  }

  /* -------------------------------------------- */
  static removeChatMessageId(messageId) {
    if (messageId) {
      game.messages.get(messageId)?.delete();
    }
  }

  static findChatMessageId(current) {
    return Imperium5Utility.getChatMessageId(Imperium5Utility.findChatMessage(current))
  }

  static getChatMessageId(node) {
    return node?.attributes.getNamedItem('data-message-id')?.value
  }

  static findChatMessage(current) {
    return Imperium5Utility.findNodeMatching(current, it => it.classList.contains('chat-message') && it.attributes.getNamedItem('data-message-id'))
  }

  static findNodeMatching(current, predicate) {
    if (current) {
      if (predicate(current)) {
        return current;
      }
      return Imperium5Utility.findNodeMatching(current.parentElement, predicate)
    }
    return undefined;
  }

  /* -------------------------------------------- */
  static templateData(it) {
    return Imperium5Utility.data(it)?.data ?? {}
  }

  /* -------------------------------------------- */
  static data(it) {
    if (it instanceof Actor || it instanceof Item || it instanceof Combatant) {
      return it.data;
    }
    return it;
  }

  /* -------------------------------------------- */
  static createDirectOptionList(min, max) {
    let options = {};
    for (let i = min; i <= max; i++) {
      options[`${i}`] = `${i}`;
    }
    return options;
  }

  /* -------------------------------------------- */
  static buildListOptions(min, max) {
    let options = ""
    for (let i = min; i <= max; i++) {
      options += `<option value="${i}">${i}</option>`
    }
    return options;
  }

  /* -------------------------------------------- */
  static getTarget() {
    if (game.user.targets && game.user.targets.size == 1) {
      for (let target of game.user.targets) {
        return target;
      }
    }
    return undefined;
  }

  /* -------------------------------------------- */
  static updateRollData(rollData) {

    let id = rollData.rollId;
    let oldRollData = this.rollDataStore[id] || {};
    let newRollData = mergeObject(oldRollData, rollData);
    this.rollDataStore[id] = newRollData;
  }
  /* -------------------------------------------- */
  static saveRollData(rollData) {
    game.socket.emit("system.pegasus-rpg", {
      name: "msg_update_roll", data: rollData
    }); // Notify all other clients of the roll    
    this.updateRollData(rollData);
  }

  /* -------------------------------------------- */
  static getRollData(id) {
    return this.rollDataStore[id];
  }

  /* -------------------------------------------- */
  static async onSocketMesssage(msg) {
    console.log("SOCKET MESSAGE", msg.name)
    if (msg.name == "msg_update_roll") {
      this.updateRollData(msg.data)
    }
  }

  /* -------------------------------------------- */
  static chatDataSetup(content, modeOverride, isRoll = false, forceWhisper) {
    let chatData = {
      user: game.user.id,
      rollMode: modeOverride || game.settings.get("core", "rollMode"),
      content: content
    };

    if (["gmroll", "blindroll"].includes(chatData.rollMode)) chatData["whisper"] = ChatMessage.getWhisperRecipients("GM").map(u => u.id)
    if (chatData.rollMode === "blindroll") chatData["blind"] = true
    else if (chatData.rollMode === "selfroll") chatData["whisper"] = [game.user]

    if (forceWhisper) { // Final force !
      chatData["speaker"] = ChatMessage.getSpeaker()
      chatData["whisper"] = ChatMessage.getWhisperRecipients(forceWhisper)
    }

    return chatData;
  }

  /* -------------------------------------------- */
  static async loadCompendiumData(compendium) {
    const pack = game.packs.get(compendium);
    return await pack?.getDocuments() ?? [];
  }

  /* -------------------------------------------- */
  static async loadCompendium(compendium, filter = item => true) {
    let compendiumData = await this.loadCompendiumData(compendium);
    //console.log("Compendium", compendiumData);
    return compendiumData.filter(filter);
  }

  /* -------------------------------------------- */
  static async showDiceSoNice(roll, rollMode) {
    if (game.modules.get("dice-so-nice")?.active) {
      if (game.dice3d) {
        let whisper = null;
        let blind = false;
        rollMode = rollMode ?? game.settings.get("core", "rollMode");
        switch (rollMode) {
          case "blindroll": //GM only
            blind = true;
          case "gmroll": //GM + rolling player
            whisper = this.getUsers(user => user.isGM);
            break;
          case "roll": //everybody
            whisper = this.getUsers(user => user.active);
            break;
          case "selfroll":
            whisper = [game.user.id];
            break;
        }
        await game.dice3d.showForRoll(roll, game.user, true, whisper, blind)
      }
    }
  }

  /* -------------------------------------------- */
  static async rollImperium5(rollData) {

    let dicePool = [{ name: "stat", level: 0, statmod: 0 }, { name: "spec", level: 0 }, { name: "bonus", level: 0 }, { name: "hindrance", level: 0 }, { name: "other", level: 0 }];
    if (rollData.stat) {
      dicePool[0].level += Number(rollData.stat.value);
      dicePool[0].statmod = Number(rollData.stat.mod);
    }
    if (rollData.statDicesLevel) {
      dicePool[0].level = rollData.statDicesLevel;
    }
    if (rollData.selectedSpec && rollData.selectedSpec != "0") {
      rollData.spec = rollData.specList.find(item => item._id == rollData.selectedSpec);
      rollData.spec.data.dice = Imperium5Utility.getDiceFromLevel(rollData.spec.data.level);
    }
    if (rollData.spec) {
      dicePool[1].level += Number(rollData.spec.data.level);
    }
    if (rollData.specDicesLevel) {
      dicePool[1].level = rollData.specDicesLevel;
    }
    if (rollData.bonusDicesLevel) {
      dicePool[2].level += Number(rollData.bonusDicesLevel);
    }
    if (rollData.hindranceDicesLevel) {
      dicePool[3].level += Number(rollData.hindranceDicesLevel);
    }
    if (rollData.otherDicesLevel) {
      dicePool[4].level += Number(rollData.otherDicesLevel);
    }

    let diceFormulaTab = [];
    for (let diceGroup of dicePool) {
      diceFormulaTab.push(this.getFoundryDiceFromLevel(diceGroup.level))
    }
    let diceFormula = '{' + diceFormulaTab.join(', ') + '}kh';

    // Performs roll
    let myRoll = rollData.roll;
    if (!myRoll) { // New rolls only of no rerolls
      myRoll = new Roll(diceFormula).roll({ async: false });
      console.log("ROLL : ", diceFormula)
      await this.showDiceSoNice(myRoll, game.settings.get("core", "rollMode"));
      rollData.roll = myRoll
    }

    // Final score and keep data
    rollData.finalScore = myRoll.total + dicePool[0].statmod;

    if (rollData.damages) {
      let dmgFormula = this.getFoundryDiceFromLevel(rollData.damages.value)
      let dmgRoll = new Roll(dmgFormula).roll({ async: false });
      await this.showDiceSoNice(dmgRoll, game.settings.get("core", "rollMode"));
      rollData.dmgResult = dmgRoll.total;
    }

    this.createChatWithRollMode(rollData.alias, {
      content: await renderTemplate(`systems/fvtt-imperium5/templates/chat-generic-result.html`, rollData)
    });

    // Init stuf
    if (rollData.isInit) {
      let combat = game.combats.get(rollData.combatId)
      combat.updateEmbeddedDocuments("Combatant", [{ _id: rollData.combatantId, initiative: rollData.finalScore }]);
    }

    //this.removeUsedPerkEffects( rollData) // Unused for now

    // And save the roll
    this.saveRollData(rollData);
  }

  /* -------------------------  ------------------- */
  static async updateRoll(rollData) {

    let diceResults = rollData.diceResults
    let sortedRoll = []
    for (let i = 0; i < 10; i++) {
      sortedRoll[i] = 0;
    }
    for (let dice of diceResults) {
      sortedRoll[dice.result]++
    }
    let index = 0;
    let bestRoll = 0;
    for (let i = 0; i < 10; i++) {
      if (sortedRoll[i] > bestRoll) {
        bestRoll = sortedRoll[i]
        index = i
      }
    }
    let bestScore = (bestRoll * 10) + index
    rollData.bestScore = bestScore
    rollData.finalScore = bestScore + rollData.negativeModifier + rollData.positiveModifier

    this.saveRollData(rollData)

    this.createChatWithRollMode(rollData.alias, {
      content: await renderTemplate(`systems/fvtt-weapons-of-the-gods/templates/chat-generic-result.html`, rollData)
    });
  }

  /* -------------------------  ------------------- */
  static async rerollDice(actorId, diceIndex = -1) {
    let actor = game.actors.get(actorId)
    let rollData = actor.getRollData()

  }

  /* -------------------------------------------- */
  static getUsers(filter) {
    return game.users.filter(filter).map(user => user.data._id)
  }
  /* -------------------------------------------- */
  static getWhisperRecipients(rollMode, name) {
    switch (rollMode) {
      case "blindroll": return this.getUsers(user => user.isGM)
      case "gmroll": return this.getWhisperRecipientsAndGMs(name)
      case "selfroll": return [game.user.id]
    }
    return undefined;
  }
  /* -------------------------------------------- */
  static getWhisperRecipientsAndGMs(name) {
    let recep1 = ChatMessage.getWhisperRecipients(name) || [];
    return recep1.concat(ChatMessage.getWhisperRecipients('GM'));
  }

  /* -------------------------------------------- */
  static blindMessageToGM(chatOptions) {
    let chatGM = duplicate(chatOptions);
    chatGM.whisper = this.getUsers(user => user.isGM);
    chatGM.content = "Message en aveugle à " + game.user.name + "<br>" + chatOptions.content;
    console.log("blindMessageToGM", chatGM);
    game.socket.emit("system.fvtt-pegasus-rgp", { msg: "msg_gm_chat_message", data: chatGM });
  }


  /* -------------------------------------------- */
  static async searchItem(dataItem) {
    let item
    if (dataItem.pack) {
      item = await fromUuid("Compendium." + dataItem.pack + "." + dataItem.id)
    } else {
      item = game.items.get(dataItem.id)
    } 
    return item
  }

  /* -------------------------------------------- */
  static split3Columns(data) {

    let array = [[], [], []];
    if (data == undefined) return array;

    let col = 0;
    for (let key in data) {
      let keyword = data[key];
      keyword.key = key; // Self-reference
      array[col].push(keyword);
      col++;
      if (col == 3) col = 0;
    }
    return array;
  }

  /* -------------------------------------------- */
  static createChatMessage(name, rollMode, chatOptions) {
    switch (rollMode) {
      case "blindroll": // GM only
        if (!game.user.isGM) {
          this.blindMessageToGM(chatOptions)

          chatOptions.whisper = [game.user.id];
          chatOptions.content = "Message only to the GM"
        }
        else {
          chatOptions.whisper = this.getUsers(user => user.isGM)
        }
        break;
      default:
        chatOptions.whisper = this.getWhisperRecipients(rollMode, name)
        break;
    }
    chatOptions.alias = chatOptions.alias || name;
    ChatMessage.create(chatOptions);
  }

  /* -------------------------------------------- */
  static getBasicRollData() {
    let rollData = {
      rollId: randomID(16),
      rollMode: game.settings.get("core", "rollMode"),
      effectsList: [],
      armorsList: [],
      weaponsList: [],
      equipmentsList: [],
      optionsDiceList: Imperium5Utility.getOptionsDiceList()
    }
    Imperium5Utility.updateWithTarget(rollData)
    return rollData
  }

  /* -------------------------------------------- */
  static updateWithTarget(rollData) {
    let objectDefender
    let target = Imperium5Utility.getTarget()
    if (target) {
      let defenderActor = game.actors.get(target.data.actorId)
      objectDefender = Imperium5Utility.data(defenderActor)
      objectDefender = mergeObject(objectDefender, target.data.actorData)
      rollData.defender = objectDefender
      rollData.attackerId = this.id
      rollData.defenderId = objectDefender._id
    }
  }

  /* -------------------------------------------- */
  static createChatWithRollMode(name, chatOptions) {
    this.createChatMessage(name, game.settings.get("core", "rollMode"), chatOptions)
  }

  /* -------------------------------------------- */
  static async confirmDelete(actorSheet, li) {
    let itemId = li.data("item-id");
    let msgTxt = "<p>Etes vous certain de vouloir supprimer cet objet ?";
    let buttons = {
      delete: {
        icon: '<i class="fas fa-check"></i>',
        label: "Oui",
        callback: () => {
          actorSheet.actor.deleteEmbeddedDocuments("Item", [itemId]);
          li.slideUp(200, () => actorSheet.render(false));
        }
      },
      cancel: {
        icon: '<i class="fas fa-times"></i>',
        label: "Annuler"
      }
    }
    msgTxt += "</p>";
    let d = new Dialog({
      title: "Confirmer la suppression",
      content: msgTxt,
      buttons: buttons,
      default: "cancel"
    });
    d.render(true);
  }

}